#include <stdio.h>

#include "lua.h"
#include "lualib.h"
//#include "lauxlib.h"
#include "tolua.h"

#include "idex.h"
#include "hash.h"

#include "main.h"
#include "api.h"

Player *pplayer;
Unit *punit;
City *pcity;

lua_CFunction hook;


/*
static void event_handler(const char *event_name, lua_Function c_handler)
{
}
*/
void add_event_handler(const char *event_name, const char *c_handler)
{
  printf("add_event_handler: %s\n", event_name);
}

Player *get_player(void)
{
  return pplayer;
}

Unit *get_unit(void)
{
  return punit;
}

void dude(int number)
{
  printf("DUDE: %d\n", number);
}

void bar(Player *player)
{
    printf("BAR: %s\n", player->name);
}

struct {
  lua_State *state;

  struct hash_table *signals;
  struct hash_table *callbacks;
} script_data;

static bool callback_status_get(const char *callback_name)
{
  struct hash_table *hash = script_data.callbacks;

  return (bool) hash_lookup_data(hash, callback_name);
}

static bool callback_status_set(const char *callback_name, bool enabled)
{
  struct hash_table *hash = script_data.callbacks;
  const void *key;

  if (hash_lookup(hash, callback_name, &key, NULL)) {
    hash_replace(hash, key, (void *) enabled);
    return FALSE;
  } else {
    hash_insert(hash, callback_name, (void *) enabled);
    return TRUE;
  }
}

bool callback_invoke_valist(const char *callback_name,
			    va_list args)
{
  lua_State *state = script_data.state;

  if (callback_status_get(callback_name)) {
    int nargs;
    bool res;

    /* the function name */
    lua_getglobal(state, callback_name);

    nargs = 0;

    /* call the function with nargs arguments, return 1 results */
    lua_call(state, nargs, 1);

    res = lua_toboolean(state, -1);
    lua_pop(state, 1);

    if (res) {
      callback_status_set(callback_name, FALSE);
      return TRUE;
    }
  }
  return FALSE;
}

/* get 'struct signal_list' and related functions: */
#define SPECLIST_TAG signal
#define SPECLIST_TYPE char
#include "speclist.h"

#define signal_list_iterate(signallist, psignal) \
    TYPED_LIST_ITERATE(char, signallist, psignal)
#define signal_list_iterate_end  LIST_ITERATE_END

static void script_signal_invoke_valist(const char *signal_name, va_list args)
{
  struct hash_table *hash;
  struct signal_list *signals;

  hash = script_data.signals;
  signals = hash_lookup_data(hash, signal_name);

  if (signals) {
    signal_list_iterate(signals, callback_name) {
      if (callback_invoke_valist(callback_name, args)) {
	signal_list_unlink(signals, callback_name);
      }
    } signal_list_iterate_end;
  }
}

void script_signal_emit_valist(const char *signal_name, va_list args)
{
  script_signal_invoke_valist(signal_name, args);
}

void script_signal_new_valist(const char *signal_name, va_list args)
{
  struct hash_table *hash;
  struct signal_list *list;
  char *name;

  hash = script_data.signals;
  list = signal_list_new();
  name = strdup(signal_name);

  if (!hash_insert(hash, name, list)) {
    signal_list_free(list);
    free(name);
  }
}

void api_signal_connect(const char *signal_name, const char *callback_name)
{
  struct signal_list *list;
  char *name;

  list = hash_lookup_data(script_data.signals, signal_name);
  name = strdup(callback_name);

  signal_list_append(list, name);
  if (!callback_status_set(name, TRUE)) {
    free(name);
  }
}

bool script_init(void)
{
  lua_State *state;

  state = lua_open();
  if (!state) {
    return FALSE;
  }

  script_data.state = state;

  luaopen_base(state);
  luaopen_string(state);
  luaopen_io(state);
  luaopen_debug(state);
  luaopen_table(state);

  tolua_api_open(state);

  api_store_init();

  script_data.signals = hash_new(hash_fval_string, hash_fcmp_string);
  script_data.callbacks = hash_new(hash_fval_string, hash_fcmp_string);

  return TRUE;
}

void script_free(void)
{
  hash_free(script_data.signals);
  hash_free(script_data.callbacks);

  api_store_free();
}

bool script_do_file(const char *filename)
{
  return lua_dofile(script_data.state, filename);
}

int main(int argc, char *argv[])
{
  va_list ap;

  if (argc < 2) {
    return 1;
  }

  idex_init();

  script_init();

  pplayer = malloc(sizeof(*pplayer));
  pplayer->player_no = 10;
  strcpy(pplayer->name, "Toto");

  punit = malloc(sizeof(*punit));
  punit->id = 69;
  punit->owner = 0;
  punit->homecity = 6969;
  idex_register_unit(punit);

  pcity = malloc(sizeof(*pcity));
  strcpy(pcity->name, "Kansas");
  pcity->id = 6969;
  pcity->owner = 0;
  idex_register_city(pcity);

  script_signal_new_valist("turn", ap);

  script_do_file(argv[1]);

  script_signal_emit_valist("turn", ap);
  script_signal_emit_valist("turn", ap);

  script_free();

  idex_free();
  return 0;
}

